"""
Trains a neural network to fit the opacity data.
"""
import pickle

import numpy as np

try:
    import keras
    import tensorflow as tf
    from tensorflow.keras import layers
except ModuleNotFoundError(
    "The optional neural network-related packages have not been installed. If you would like to"
    + "use cortecs to fit with a neural network, please install cortecs with neural network functionality"
    + "as follows: python3 -m pip install -e .[neural_networks]"
):
    pass

from cortecs.fit.metrics import *


# todo: loop over all wavelengths.
def prep_neural_net(cross_section):
    pass


def unravel_data(x, y, z=None, tileboth=False):
    """
    unravels the data into a single column. takes log the log of quantities as well.

    todo: move to a utils file?

    Inputs
    ------
    x: array-like
        first dimension of data.
    y: array-like
        second dimension of data.
    z: array-like
        third dimension of data.
    """

    if isinstance(z, type(None)):
        if tileboth:
            return np.tile(np.log10(x), len(y))
        return np.repeat(np.log10(x), len(y))
    if tileboth:
        return np.tile(np.tile(np.log10(x), len(y)), len(z))

    return np.tile(np.repeat(np.log10(x), len(y)), len(z))


def fit_neural_net(
    cross_section,
    P,
    T,
    wl,
    n_layers=3,
    n_neurons=8,
    activation="sigmoid",
    learn_rate=0.04,
    loss="mean_squared_error",
    epochs=2000,
    verbose=0,
    sequential_model=None,
    plot=False,
):
    """
    trains a neural network to fit the opacity data.

    Inputs
    -------
        :cross_section: (ntemp x npressure) the array of cross-sections being fit.
        :P: (array) pressure grid corresponding to the cross-sections.
        :T: (array) temperature grid corresponding to the cross-sections.
        :wl: (array) wavelength grid corresponding to the cross-sections.
        :n_layers: (int) number of layers in the neural network. Increasing this number increases the flexibility
                of the model, but it also increases the number of parameters that need to be fit — and hence the model
                size and training time.
        :n_neurons: (int) number of neurons in each layer. Increasing this number increases the flexibility
                of the model, but it also increases the number of parameters that need to be fit — and hence the model
                size and training time.
        :activation: (str) activation function to use. See the keras documentation for more information:
                https://keras.io/api/layers/activations/
        :learn_rate: (float) learning rate for the optimizer. Increasing this number can help the model converge faster,
                but it can also cause the model to diverge.
        :loss: (str) loss function to use. See the keras documentation for more information:
                https://keras.io/api/losses/
        :epochs: (int) number of epochs to train for. Increasing this number can help the model converge better,
                but it primarily makes the training time longer.
        :verbose: (int) verbosity of the training.
        :sequential_model: (keras.Sequential) if not None, this is the neural network to be trained.
        :plot: (bool) whether to plot the loss.


    Returns
    -------
        :history: the history object from the keras fit method
        :neural_network: the trained neural network

    todo: implement all_wl
    """

    if sequential_model is None:
        layers_list = []
        for i in range(n_layers):
            layers_list += [layers.Dense(n_neurons, activation=activation)]

        layers_list += [layers.Dense(1)]  # final layer to predict single value

        neural_network = keras.Sequential(layers_list)

    else:
        neural_network = sequential_model

    # running the legacy Adam optimizer for the Mac folks!
    neural_network.compile(
        loss=loss, optimizer=tf.keras.optimizers.legacy.Adam(learn_rate)
    )

    # unpack opacity data. todo: not repeat this for every wavelength!
    P_unraveled = unravel_data(P, T, wl)
    T_unraveled = unravel_data(T, P, wl, tileboth=True)

    # todo: try to predict everything in one big net, not wavelength-independent?
    input_array = np.column_stack([P_unraveled, T_unraveled])
    predict_data_flattened = cross_section.flatten()

    history = neural_network.fit(
        input_array,
        predict_data_flattened,  # totally fine to overfit! we're not extrapolating, nor are we interpolating.
        verbose=verbose,
        epochs=epochs,
    )

    if plot:
        plot_loss(history)

    return history, neural_network


def save_neural_net(filename, fit_results):
    """
    saves the neural network to a file.


    Inputs
    -------
        :filename: (str) filename to save the neural network to.
        :fit_results: (keras.Sequential) the neural network to save.


    Returns
    -------
        :None:
    """
    neural_network = fit_results[1]
    n_layers = len(neural_network.layers)
    all_weights = []
    all_biases = []

    for layer in range(n_layers):
        layer_weights = neural_network.layers[layer].get_weights()[0]
        layer_biases = neural_network.layers[layer].get_weights()[1]
        all_weights += [layer_weights]
        all_biases += [layer_biases]

    # save all_weights and all_biases to a pickle
    with open(filename + ".pkl", "wb") as f:
        pickle.dump([all_weights, all_biases], f)

    # now, let's read it back in and make sure it's the same.

    return



"""
Reads opacity data from various sources.

author: @arjunsavel
"""
import pickle

import h5py
import numpy as np
import pandas as pd
from tqdm import tqdm

AMU = 1.6605390666e-24  # atomic mass unit in cgs. From astropy!


class loader_base(object):
    """
    loads in opacity data from various sources. To be passed on to Opac object.

    todo: tutorial on how to use this?
    """

    def __init__(
        self,
        wl_key="wno",
        T_key="T",
        P_key="P",
        cross_section_key="xsec",
        wl_style="wno",
        temperature_axis=0,
        pressure_axis=1,
        wavelength_axis=2,
    ):
        """
        sets the keys for the loader object.
        """
        self.wl_key = wl_key
        self.T_key = T_key
        self.P_key = P_key
        self.cross_section_key = cross_section_key
        self.wl_style = wl_style
        self.temperature_axis = temperature_axis
        self.pressure_axis = pressure_axis
        self.wavelength_axis = wavelength_axis

    def load(self, filename):
        """
        loads in opacity data from various sources. To be passed on to Opac object.

        Inputs
        ------
        filename : str
            name of file to load

        Outputs
        -------
        wl : np.ndarray
            wavelengths
        T : np.ndarray
            temperatures
        P : np.ndarray
            pressures
        cross_section : np.ndarray
            cross sections for the opacity
        """

        hf = h5py.File(filename, "r")
        wl = np.array(hf[self.wl_key], dtype=np.float64)
        T = np.array(hf[self.T_key], dtype=np.float64)
        P = np.array(hf[self.P_key], dtype=np.float64)
        cross_section = np.array(hf[self.cross_section_key], dtype=np.float64)
        hf.close()

        # want temperature index 0, pressure to 1, wavelength to 2 for standard usage.
        cross_section = np.moveaxis(
            cross_section,
            [self.temperature_axis, self.pressure_axis, self.wavelength_axis],
            [0, 1, 2],
        )

        if self.wl_style == "wno":
            wl = 1e4 / wl

        if np.all(np.diff(wl)) < 0:
            wl = wl[::-1]
            cross_section = cross_section[:, :, ::-1]
            # reverse!
        return wl, T, P, cross_section


class loader_chimera(loader_base):
    """
    loads in opacity data that are produced with the CHIMERA code.
    """


class loader_helios(loader_base):
    """
    loads in opacity data that are produced with the HELIOS ktable function.
    """

    wl_key = "wavelengths"
    T_key = "temperatures"
    P_key = "pressures"
    cross_section_key = "opacities"
    wl_style = "wl"


class loader_platon(loader_base):
    """
    loads in opacity data that are used with the PLATON code.
    """

    # atomic weights of molecules
    species_weight_dict = {
        "CO": 28.01,
        "H2O": 18.015,
        "CH4": 16.04,
        "NH3": 17.031,
        "CO2": 44.009,
        "HCN": 27.026,
        "C2H2": 26.038,
        "H2S": 34.08,
        "PH3": 33.997,
        "VO": 66.94,
        "TiO": 63.866,
        "Na": 22.99,
        "K": 39.098,
        "FeH": 55.845,
        "H2": 2.016,
        "He": 4.003,
        "H-": 1.008,
        "H": 1.008,
        "He+": 4.003,
        "H+": 1.008,
        "e-": 0.00054858,
        "H2+": 2.016,
        "H2-": 2.016,
        "H3+": 3.024,
        "H3-": 3.024,
        "H2O+": 18.015,
        "H2O-": 18.015,
        "CH4+": 16.04,
        "CH4-": 16.04,
        "C2H4": 28.054,
    }

    def load(self, cross_sec_filename, T_filename="", P_filename="", wl_filename=""):
        """
        loads in opacity data that's built for PLATON. To be passed on to Opac object.

        The temperature grid, pressure grid, and wavelength grid are saved as separate files for PLATON.

        Inputs
        ------
        filename : str
            name of file to load
        cross_sec_filename : str
            name of cross section file
        T_filename : str
            name of temperature file
        P_filename : str
            name of pressure file
        wl_filename : str
            name of wavelength file
        """
        # todo: check wl units. They're in meters here.
        # temperatures are in K.
        # pressures are in Pa, I believe.
        wl = np.load(wl_filename, allow_pickle=True)
        T = np.load(T_filename, allow_pickle=True)
        P = np.load(P_filename, allow_pickle=True)
        cross_section = np.load(
            cross_sec_filename, allow_pickle=True
        )  # packaged as T x P x wl. todo: check packing
        # cross-section units for fitting...? keep the same internally.
        species = cross_sec_filename.split("_")[-1].split(".")[0]
        try:
            self.species_weight = self.species_weight_dict[species]
        except KeyError:
            raise KeyError(
                f"Species {species} read from filename and not found in species_weight_dict. Please add it."
            )
        cross_section = cross_section * AMU * self.species_weight * 1e-4

        # set a floor to the opacities, as per the floor used in the PLATON code's opacity files.
        cross_section[np.less(cross_section, 1e-104)] = 1e-104

        # and now make it log10
        cross_section = np.log10(cross_section)

        return wl, T, P, cross_section


class loader_platon_cia(loader_base):
    """
    loads in opacity data that are used with the PLATON code's collision-induced absorption..
    """

    def load(self, cross_sec_filename, T_filename, wl_filename, species_name):
        """
        loads in opacity data that's built for PLATON. To be passed on to Opac object.

        The temperature grid, pressure grid, and wavelength grid are saved as separate files for PLATON.

        Inputs
        ------
        filename : str
            name of file to load
        cross_sec_filename : str
            name of cross section file
        T_filename : str
            name of temperature file
        wl_filename : str
            name of wavelength file
        species_name : tuples
            name of two colliding species. E.g., ('H2', 'CH4'). todo: all at once?
        """
        # todo: check wl units. They're in meters here.
        # temperatures are in K.
        # pressures are in Pa, I believe.
        wl = np.load(wl_filename, allow_pickle=True)
        T = np.load(T_filename, allow_pickle=True)
        data = pickle.load(
            open(cross_sec_filename, "rb"), encoding="latin1"
        )  # packaged as T x P x wl. todo: check packing
        cross_section = data[species_name]
        return wl, T, cross_section


class loader_exotransmit_cia(loader_base):
    """
    loads in opacity data that are used with the PLATON code's collision-induced absorption..
    """

    def check_line_break(self, line, temperature):
        """
        Checks whether the given line should be skipped

        """
        truth_val = False
        if not line or line == "\n":
            truth_val = True
        if len(line.split()) == 1 and line != "\n":  # this is a new temperature
            temperature = eval(line[:-1])
            truth_val = True
        return truth_val, temperature

    def load(self, cross_sec_filename, verbose=False):
        """
        loads in opacity data that's built for exo-transmit. To be passed on to Opac object.


        Inputs
        ------
            :CIA_file: (str) path to CIA file. e.g., 'opacCIA/opacCIA.dat'
        Outputs
        -------
            :df: (pd.DataFrame) dataframe with the CIA data.
        """

        f = open(cross_sec_filename)
        f1 = f.readlines()
        f.close()

        temperatures = []
        wavelengths = []

        species_dict = get_empty_species_dict(cross_sec_filename, verbose=verbose)

        # read through all lines in the CIA file
        temperature = 0.0  # initialize

        for line in tqdm(f1[1:], desc="Reading CIA file"):
            truth_val, temperature = self.check_line_break(line, temperature)

            if truth_val:
                continue

            values = [eval(value) for value in line.split()]
            temperatures += [temperature]
            wavelengths += [values[0]]

            for species_ind, species in enumerate(species_dict.keys()):
                species_dict[species] += [values[species_ind + 1]]

        # hm. how to do this with different numbers of species?
        df = pd.DataFrame(
            {
                "temp": temperatures,
                "wav": wavelengths,
            }
        )
        for species in species_dict.keys():
            df[species] = species_dict[species]

        columns = list(df.columns)
        columns.remove("temp")
        columns.remove("wav")

        return df.wav.values, df.temp.values, df[columns]


class loader_exotransmit(loader_base):
    """
    loads in opacity data in the exo-transmit format. This format is also used by PLATON, I believe.
    """

    def get_t_p(self, file):
        """
        Gets the temperatures and pressures of a file from its header.

        Inputs:
            :file: (str) path to file whose header we want.

        Outputs:
            header of file (string)
        """

        f = open(file)
        f1 = f.readlines()
        f.close()

        t_line = f1[0]
        p_line = f1[1]

        temperature_strings = t_line.split(" ")[:-1]
        T = np.array([eval(temp) for temp in temperature_strings])

        pressure_strings = p_line.split(" ")[:-1]
        P = np.array([eval(pres) for pres in pressure_strings])

        del f1
        return T, P

    def get_lams_and_opacities(self, file):
        """
        Takes in an opacity file and returns an array of all wavelengths within the file.

        Returns the opacities, as well — so that the file is only iterated through once.

        Inputs:
            :file: (str) path to opacity file.

        Outputs:
            :wavelengths: (numpy.array) individual wavelength points within the opacity file [m]
        """
        f = open(file)
        f1 = f.readlines()

        wavelengths = []
        opacities = []

        # read through all lines in the opacity file
        # skip through the header!
        for x in tqdm(f1[2:], desc="reading wavelengths"):
            # check if blank line
            if not x:
                continue
            # check if a wavelength line
            commad = x.replace(" ", ",")
            if len(np.array([eval(commad)]).flatten()) == 1:
                wavelengths += [eval(x[:-1])]
            else:
                # the first entry in each opacity line is the pressure
                opacity_string = x.split()[1:]
                opacity_vals = np.array([eval(opacity) for opacity in opacity_string])
                opacities += [opacity_vals]

        f.close()
        # pdb.set_trace()
        del f1
        try:
            return np.array(wavelengths), np.array(opacities)
        except:
            pdb.set_trace()

    def load(self, filename):
        """
        Loads file.

        Inputs
        ------
        filename : str
            name of file to load

        Outputs
        -------
        wl : np.ndarray
            wavelengths
        T : np.ndarray
            temperatures
        P : np.ndarray
            pressures
        """

        wl, opacities = self.get_lams_and_opacities(filename)
        T, P = self.get_t_p(filename)

        return wl, T, P, opacities


# todo: put these in the opac class?
def get_empty_species_dict(CIA_file, verbose=False):
    """
    returns a species dictioanry given a CIA file

    Inputs
    -------
        :CIA_file: (str) path to CIA file. e.g., 'opacCIA/opacCIA.dat'
        :verbose: (bool) whether to print out the species that are likely in the file.

    Outputs
    -------
        :species_dict: (dict) dictionary of species.

    """
    n_species = get_n_species(CIA_file, verbose=verbose)
    if n_species == 8:
        # todo: refactor. has to be a cleaner way to do this! infer the columns, etc.
        Hels = []
        HeHs = []
        CH4CH4s = []
        H2Hes = []
        H2CH4s = []
        H2Hs = []
        H2H2s = []
        CO2CO2s = []

        # python > 3.6 has ordered dictionaries!
        species_dict = {
            "Hels": Hels,
            "HeHs": HeHs,
            "CH4CH4s": CH4CH4s,
            "H2Hes": H2Hes,
            "H2CH4s": H2CH4s,
            "H2Hs": H2Hs,
            "H2H2s": H2H2s,
            "CO2CO2s": CO2CO2s,
        }

    elif n_species == 14:
        H2H2s = []
        H2Hes = []
        H2Hs = []
        H2CH4s = []
        CH4Ar = []
        CH4CH4s = []
        CO2CO2s = []
        HeHs = []
        N2CH4s = []
        N2H2s = []
        N2N2s = []
        O2CO2s = []
        O2N2s = []
        O2O2s = []
        species_dict = {
            "H2H2s": H2H2s,
            "H2Hes": H2Hes,
            "H2Hs": H2Hs,
            "H2CH4s": H2CH4s,
            "CH4Ar": CH4Ar,
            "CH4CH4s": CH4CH4s,
            "CO2CO2s": CO2CO2s,
            "HeHs": HeHs,
            "N2CH4s": N2CH4s,
            "N2H2s": N2H2s,
            "N2N2s": N2N2s,
            "O2CO2s": O2CO2s,
            "O2N2s": O2N2s,
            "O2O2s": O2O2s,
        }
    else:
        print("Number of species in CIA file not recognized. Check the file!")
        species_dict = {}
        species_keys = np.arange(n_species)
        for species_key in species_keys:
            species_dict[species_key] = []
    return species_dict


def get_n_species(CIA_file, verbose=False):
    """
    Returns the number of species in a CIA file.
    Inputs
    -------
        :CIA_file: (str) path to CIA file. e.g., 'opacCIA/opacCIA.dat'
    Outputs
    -------
        :n_species: (int) number of species in the CIA file.
    """

    f = open(CIA_file)
    f1 = f.readlines()
    f.close()
    # the third line fshould be the first normal one.

    line = f1[3]
    n_species = len(line.split()[1:])

    if verbose:
        print(f"Number of species in CIA file {CIA_file}: {n_species}")
        if n_species == 8:
            print(
                "Species are likely: H-el, He-H, CH4-CH4, H2-He, H2-CH4, H2-H, H2-H2, CO2-CO2"
            )
        elif n_species == 14:
            print(
                "Species are likely:  H2-H2, H2-He, H2-H, H2-CH4, CH4-Ar, CH4-CH4, CO2-CO2, He-H, N2-CH4, N2-H2, N2-N2, O2-CO2, O2-N2, O2-O2)"
            )
    del f1
    return n_species


# todo: write a writer for each of these.


class writer_base(object):
    def __init__(self):
        """
        nothing to do here
        """
        pass


class writer_exotransmit_cia(writer_base):
    """
    does writing for the CIA object, takng in an opac object.
    """

    def write(self, opac, outfile, verbose=False):
        """
        loads in opacity data that's built for exo-transmit. To be passed on to Opac object.


        Inputs
        ------
            :CIA_file: (str) path to CIA file. e.g., 'opacCIA/opacCIA.dat'
        Outputs
        -------
            :df: (pd.DataFrame) dataframe with the CIA data.
        """

        new_string = []

        # don't want to write temp and wav
        columns = list(opac.cross_section.columns)
        columns.remove("temp")
        columns.remove("wav")

        self.species_dict_interped = opac.cross_section[columns].to_dict()
        self.interped_temps = opac.T
        self.interped_wavelengths = opac.wl

        # this is pretty gross below : (
        reference_species = self.species_dict_interped[
            list(self.species_dict_interped.keys())[0]
        ]

        # todo: include the different temperature range on which to interpolate.

        self.buffer = "   "  # there's a set of spaces between each string!
        temp = 0.0  # initial value
        for i in tqdm(range(len(reference_species)), desc="Writing file"):
            new_string, temp = self.append_line_string(
                new_string,
                i,
                temp,
            )

        # todo: check the insert. and can pull wavelength grid.
        temp_string = (
            " ".join(str(temp) for temp in np.sort(np.unique(self.interped_temps)))
            + " \n"
        )
        new_string.insert(0, temp_string)

        f2 = open(outfile, "w")
        f2.writelines(new_string)
        f2.close()

    def append_line_string(
        self,
        new_string,
        i,
        temp,
    ):
        # the first line gets different treatment!
        if i == 0:
            temp = np.min(
                self.interped_temps
            )  # add the LOWEST temperature in the temperature grid!
            new_string += ["{:.12e}".format(temp) + "\n"]
        if self.interped_temps[i] != temp:
            temp = self.interped_temps[i]
            new_string += ["{:.12e}".format(temp) + "\n"]
        wavelength_string = "{:.12e}".format(self.interped_wavelengths[i])

        line_string = wavelength_string + self.buffer

        for species_key in self.species_dict_interped.keys():
            # again, this works because python dicts are ordered in 3.6+
            line_string += (
                "{:.12e}".format(
                    list(self.species_dict_interped[species_key].values())[i]
                )
                + self.buffer
            )

        new_string += [line_string + "\n"]

        return new_string, temp

    # todo: maybe the loader objects should also take an opac object. for parallel structure : )



"""
A few functions that can be used to chunk opacity files. Assumes that
opacity files are in the RT code format, that each opacity file to
be chunked is in its on folder, and that we are currently in that folder.


Example instructions / workflow:

>>> file = 'opacTiO.dat'
>>> chunk_wavelengths(file, wav_per_chunk=2598)
>>> filename = 'opacTiO'
>>> add_overlap(filename)

And that should work!

There's some replicated (and likely unnecessary) code, but it hopefully shouldn't
be too confusing. Furthermore, these functions have not been subjected to robust
unit testing, so they might not work right out of the box. Please let me know if
that's the case, and I'd be happy to debug :)

author: @arjunsavel

so far, only works for exo-transmit chunks. TODO: generalize for others.
"""

import os
from glob import glob

import numpy as np
from tqdm import tqdm

from cortecs.opac.opac import *


def chunk_wavelengths(
    file,
    nchunks=None,
    wav_per_chunk=None,
    adjust_wavelengths=False,
    loader="exotransmit",
):
    """
    Performs wavelength-chunking.
    todo: yell if the chunked results are already in the directory?

    Inputs
    -------
        :file: path to file to be chunked. e.g., 'opacFe.dat'.

        :nchunks: (int or None) Number of chunks to use, splitting the opacity file
                        into roughly even chunks. If None, wav_per_chunk must be specified.

        :wav_per_chunk: (int or None) Number of wavelengths per chunk. If None,
                        nchunks must be specified.

        :adjust_wavelengths: (bool) whether or not to scale from CGS to MKS. For most situations,
                        can be kept false.

    Outputs
    -------
        None

    Side effects
    -------------
        Creates a number of files in same directory as file, titled file*.dat.
    """
    if nchunks and wav_per_chunk:
        print(
            "Both nchunks and wav_per_chunk specified. Will default to specified wav_per_chunk."
        )
    if not nchunks and not wav_per_chunk:
        raise ValueError(
            "Cannot set nchunks and wav_per_chunk to None! One must be specified."
        )

    if not wav_per_chunk:
        opac = Opac(file, loader=loader)
        num_wavelengths = len(opac.wl)
        del opac  # clean it up
        wav_per_chunk = round(num_wavelengths / nchunks)

    header = get_header(file)

    # now get chunks

    f = open(file)
    f1 = f.readlines()

    ticker = 0
    file_suffix = 0

    # read through all lines in the opacity file. todo: from already read in opacity?
    # past the header
    write_to_file(header, file, file_suffix)
    for x in tqdm(f1[2:]):
        # print(x)
        if not x:
            continue
        commad = x.replace(" ", ",")
        # pdb.set_trace()
        if len(np.array([eval(commad)]).flatten()) == 1:  # if a wavelength line
            ticker += 1
            print(x)
            # print(file_suffix)
        # elif len(x.split(" ")) == 48 and adjust_wavelengths:  # this is ntemp, I believe
        #     x = adjust_wavelength_unit(x, 1e-4, style="full")
        # #                pass # don't need to adust wavelengths anymore!
        if ticker == wav_per_chunk + 1:
            file_suffix += 1  # start writing to different file

            ticker = 0
            write_to_file(header, file, file_suffix)

        write_to_file(x, file, file_suffix)

    f.close()
    return


def write_to_file(line, file, file_suffix, numfiles=None):
    """
    Writes (appends, really) a line to a file.

    Inputs:
        :line: (str) line to write to file.
        :file: (str) base path to file to be written to. e.g., 'opacFe'
        :file_suffix: (int) suffix to add, usually chunk number. e.g. 5

    Outputs:
        None

    Side effects:
        Writes a line to a file!
    """
    if not isinstance(numfiles, type(None)):
        file_suffix = (file_suffix) % numfiles

    true_filename = f"{file[:-4] + str(file_suffix) + '.dat'}"
    f = open(true_filename, "a")
    f.write(line)
    f.close()


def get_header(file):
    """
    Gets the header of a file.

    Inputs:
        :file: (str) path to file whose header we want.

    Outputs:
        header of file (string)
    """

    f = open(file)
    f1 = f.readlines()
    f.close()

    return f1[0] + f1[1]


def add_lams(max_lam_to_add_ind, file, next_file):
    """
    Takes the first `max_lam_to_add_ind` opacities from next_file and appends them to file.

    Inputs:
        :max_lam_to_add_ind: (int) number of wavelength points to add from one file to the other.
        :file: (str) (str) path to file to which wavelength points are being *added*.
        :next_file: (str) path to file from which wavelength points are being drawn.

    Outputs:
        None

    Side effects:
        Modifies file.
    """
    try:
        f = open(next_file)
    except FileNotFoundError:
        print(f"{next_file} not found. Moving on!")
        return
    f1 = f.readlines()[2:]  # first two files of opacity are header info
    f.close()
    if max_lam_to_add_ind >= len(f1):
        raise IndexError("Try choosing a larger chunk size.")

    ticker = 0

    # read through all lines in the opacity file
    for x in f1:
        # skip blank lines
        if not x:
            continue

        # check if wavelength line
        commad = x.replace(" ", ",")
        if len(np.array([eval(commad)]).flatten()) == 1:
            ticker += 1

        # append to file
        f2 = open(file, "a")
        f2.write(x)
        f2.close()
        if ticker == max_lam_to_add_ind:
            return


def add_overlap(filename, v_max=11463.5):
    """
    Adds overlap from file n+1 to file n. The last file has nothing added to it. This
    step is necessary for the Doppler-on version of the RT code.

    ***Assumes that the current directory only has files labeled 'filename*.dat'***

    Inputs:
        :filename: (str) "base name" of the opacity chunks. e.g., 'opacFe', corresponding to
                        'opacFe*.dat'.

        :v_max: (float, default 11463.5) maximum velocity that will be Doppler-shifted to in the
                        RT code. Twice this value is used to calculate how much overlap to include
                        (+ an extra 20 wavelength points, to be safe!). The default comes from Tad's
                        WASP-76b GCM outputs. [m/s]


    Output:
        None

    Side effects:
        Modifies every 'filename*.dat' file.
    """
    print("for sure adding overlap")
    files = glob(filename + "*.dat")
    for i, file in tqdm(
        enumerate(files), total=len(files), position=0, leave=True, desc="eeeeee"
    ):  # don't include the last file
        if file == filename + ".dat":
            continue  # only add over lap to the chunked file
        print("for sure adding overlapppp")

        next_file = filename + str(i + 1) + ".dat"

        # go into file n to determine the delta lambda
        opac = Opac(file, loader="exotransmit")
        curr_lams = opac.wl
        del opac

        try:
            opac = Opac(next_file, loader="exotransmit")
            next_lams = opac.wl
            del opac
        except FileNotFoundError:
            print(f"{next_file} not found. Moving on!")
            continue

        c = 3e8  # m/s
        max_curr_lam = np.max(curr_lams)
        delta_lam = 2 * max_curr_lam * v_max / c  # delta_lambda/lambda = v/c

        # add another 20 indices to be safe!
        # max_lam_to_add_ind = (
        #     np.argmin(np.abs(next_lams - (max_curr_lam + delta_lam))) + 20
        # )
        max_lam_to_add_ind = np.argmin(np.abs(next_lams - (max_curr_lam + delta_lam)))
        print("max lam to add")
        print(max_lam_to_add_ind)

        add_lams(max_lam_to_add_ind, file, next_file)



"""
Module for fitting polynomial functions in temperature and pressure to opacity functions.
"""

import jax.numpy as jnp
import numpy as np

# todo: loop over all wavelengths.


def prep_polynomial(cross_section, **kargs):
    pass


# @jax.jit
def fit_polynomial(
    Z,
    P,
    T,
    prep_res,
    plot=False,
):
    """
    fits a polynomial to the opacity data.

    Inputs
    -------
        :Z: (n_temp x n_pres) it's an array.
        :P: pressure grid
        :T: temperature grid
        :prep_res: (n_temp x n_pres) PCA components
        :plot: (bool) whether to plot the fit.
        :savename: (str) if not None, the PCA components will be saved to this filename.

    Returns
    -------
        :coeff: (nc x pixels) PCA coefficients
    """
    X, Y = jnp.meshgrid(jnp.log10(T), jnp.log10(P), copy=True)

    X = X.flatten()
    Y = Y.flatten()

    # took out the y**2 term

    A = jnp.array(
        [
            X * 0 + 1,
            X,
            Y,
            X**2,
            X**2 * Y,
            X**2 * Y**2,
            X * Y**2,
            X * Y,
            X * Y**3,
            Y**4,
            Y**5,
            Y**6,
            Y**7,
            Y**12,
            X * Y**9,
            np.power(X, 1 / 4),
        ]
    ).T

    B = Z.flatten()

    coeff, _, _, _ = np.linalg.lstsq(A, B, rcond=-1)

    return coeff


def save_polynomial(savename, coeff):
    """
    Saves the polynomial coefficients to a file.

    Inputs
    -------
        :savename: (str) if not None, the PCA components will be saved to this filename.
        :coeff: (nc x ntemp) PCA coefficients
    """
    np.savetxt(savename + ".txt", coeff)



# -*- coding: utf-8 -*-

__uri__ = "https://github.com/arjunsavel/cortecs"
__author__ = "Arjun B. Savel, Megan Bedell, Eliza M.-R. Kempton"
__email__ = "asavel@gmail.com"
__license__ = "MIT"
__version__ = "0.3.1"
__release__ = "0.3.1"
__description__ = "Compress opacity for radiative transfer"



"""
The high-level API for evaluating. Requires the Opac object and the Fitter object.
"""
from cortecs.eval.eval_neural_net import *
from cortecs.eval.eval_pca import *
from cortecs.eval.eval_polynomial import *
from cortecs.opac.io import *

AMU = 1.6605390666e-24  # atomic mass unit in cgs. From astropy!


class Evaluator(object):
    """
    evaluates the opacity data.
    """

    method_dict = {
        "pca": eval_pca,
        "neural_net": eval_neural_net,
        "polynomial": eval_polynomial,
    }

    def __init__(self, opac, fitter, **eval_kwargs):
        """ """
        method = fitter.method
        self.opac = opac
        self.load_obj = opac.load_obj
        self.wl = opac.wl
        self.P = opac.P
        self.T = opac.T
        self.fit_kwargs = fitter.fitter_kwargs

        if not hasattr(fitter, "fitter_results"):
            raise ValueError("Fitter must be run before being passed to an Evaluator.")

        self.fitter_results = fitter.fitter_results
        self.eval_kwargs = eval_kwargs

        self.eval_func = self.method_dict[method]

        return

    def eval(self, temperature, pressure, wavelength):
        """
        evaluates the opacity data.

        Inputs
        ------
        temperature: float
            The temperature to evaluate at.
        pressure: float
            The pressure to evaluate at.
        wavelength: float
            The wavelength to evaluate at.

        Outputs
        -------
        eval_result: float
            The evaluated opacity.
        """
        if temperature > self.T.max() or temperature < self.T.min():
            raise ValueError(
                f"Temperature {temperature} is outside the range of the data."
            )
        if pressure > self.P.max() or pressure < self.P.min():
            raise ValueError(f"Pressure {pressure} is outside the range of the data.")
        if wavelength > self.wl.max() or wavelength < self.wl.min():
            raise ValueError(
                f"Wavelength {wavelength} is outside the range of the data."
            )

        eval_result = self.eval_func(
            temperature,
            pressure,
            wavelength,
            self.T,
            self.P,
            self.wl,
            self.fitter_results,
            **self.eval_kwargs,
            **self.fit_kwargs,
        )
        # we need to recast the units if PLATON was used.
        if isinstance(self.load_obj, loader_platon):
            eval_result = 10**eval_result
            eval_result = eval_result / self.load_obj.species_weight / AMU / 1e-4
            if eval_result == 0:
                eval_result = 1e-104  # todo: fix this inf situation

        # todo: patch the eval neural net? it doesn't return the biases?

        return eval_result



"""
This file holds the classes for evaluating opacity data as trained by a neural network.

maybe call all of these eval_poly.py, eval_neural_net.py, etc.?

author: @arjunsavel
"""
import jax
import jax.numpy as jnp
from jax import lax


@jax.jit
def feed_forward_equal_layer_sizes(x, n_layers, weights, biases):
    """
    feed forward neural network. this is a function that takes in the input, weights, and biases and returns the output.

    This function only works if all layers have the same size.

    Inputs
    ------
    x: array-like
        the input to the neural network.
    n_layers: int
        the number of layers in the neural network.
    weights: list
        the weights of the neural network.
    biases: list
    """

    def inner_function(i, x):
        return jax.nn.sigmoid(x.dot(weights[i]) + biases[i])

    res = lax.fori_loop(0, n_layers, inner_function, x)

    return res


# todo: add test to make sure feed forward actually works!
def feed_forward(x, n_layers, weights, biases):
    """
    feed forward neural network. this is a function that takes in the input, weights, and biases and returns the output.


    Inputs
    ------
    x: array-like
        the input to the neural network.
    n_layers: int
        the number of layers in the neural network.
    weights: list
        the weights of the neural network.
    biases: list
    """
    res = x
    for i in range(n_layers - 1):
        res = jax.nn.sigmoid(res.dot(weights[i]) + biases[i])
    res = res.dot(weights[-1]) + biases[-1]
    return res


# @jax.jit
def eval_neural_net(
    T,
    P,
    temperatures=None,
    pressures=None,
    wavelengths=None,
    n_layers=None,
    weights=None,
    biases=None,
    **kwargs
):
    """
    evaluates the neural network at a given temperature and pressure.

    Inputs
    ------
    T: float
        The temperature to evaluate at.
    P: float
        The pressure to evaluate at.
    n_layers: int
        The number of layers in the neural network.
    weights: list
        The weights of the neural network.
    biases: list
        The biases of the neural network.
    """
    x = jnp.array([jnp.log10(T), jnp.log10(P)])
    res = feed_forward(x, n_layers, weights, biases)
    return res



import jax
import jax.numpy as jnp


@jax.jit
def eval_polynomial(
    temperature,
    pressure,
    wavelength,
    temperatures,
    pressures,
    wavelengths,
    fitter_results,
    **kwargs,
):
    """
    evaluates the polynomial at a given temperature and pressure.

    Inputs
    ------
    temperature: float
        temperature in K.
    pressure: float
        pressure in bar.
    wavelength: float
        wavelength in microns.
    temperatures: array-like
        temperature grid.
    pressures: array-like
        pressure grid.
    wavelengths: array-like
        wavelength grid.
    fitter_results: array-like
        the PCA coefficients.

    Returns
    -------
    xsec_int: float
        The cross section at the given temperature and pressure.

    """
    TT = jnp.log10(temperature)
    PP = jnp.log10(pressure)
    coeffs = fitter_results[1]

    j = jnp.argmin(jnp.abs(wavelengths - wavelength))

    z2 = (
        1.0,
        TT,
        PP,
        TT**2,
        TT**2 * PP,
        TT**2 * PP**2,
        TT * PP**2,
        TT * PP,
        TT * PP**3,
        PP**4,
        PP**5,
        PP**6,
        PP**7,
        PP**12,
        TT * PP**9,
        TT**0.25,
    )
    Nterms = len(z2)

    xsec_int = 0.0
    for k in range(Nterms):
        xsec_int += z2[k] * coeffs[j, k]

    return xsec_int



"""
Functions that can be used to interpolate CIA to a higher
resolution and subsequently chunk it up.
Example instructions / workflow:
>>> reference_file = '../opacFe/opacFe.dat'
>>> CIA_file = 'opacCIA.dat'
>>> interpolate_cia(CIA_file, reference_file)
>>> CIA_file = 'opacCIA_highres.dat'
>>> ref_file_base = '../opacFe/opacFe'
>>> chunk_wavelengths_cia(CIA_file, ref_file_base)
And that should work!

todo: generalize to PLATON CIA.

author: @arjunsavel
"""
import os
from glob import glob

import numpy as np
from tqdm import tqdm

from cortecs.opac.chunking import *
from cortecs.opac.io import *

# Pt. 1: Interpolation


def check_temp_grid(df, real_temperature_grid, cia_file):
    """
    Checks that the temperature grid of the CIA file is the same as the reference file.
    Inputs
    ------
        :df: (pd.DataFrame) dataframe with the CIA data.
        :real_temperature_grid: (np.array) temperature values for which the opacity file had been
                    computed.
    Outputs
    -------
        None
    Side effects
    -------------
        Raises a ValueError if the temperature grid is not the same.
    """

    for temp in real_temperature_grid:
        if temp not in df.temp.unique():
            print(
                f"Temperature {temp} not in CIA file {cia_file}!"
                + f" Cannot interpolate in temperature yet. Will set these values to 0."
            )
    return


def check_wav_grid(reference_opac, df):
    """
    Checks that the wavelength grid of the CIA file is the same as the reference file.
    Inputs
    ------
        :reference_opac: (Opac) Opac object with the wavelength grid of interest.
        :df: (pd.DataFrame) dataframe with the CIA data.
    Outputs
    -------
        None
    Side effects
    -------------
        Raises a ValueError if the wavelength grid is not the same.
    """
    if reference_opac.wl.max() > df.wav.max() or reference_opac.wl.min() < df.wav.min():
        print(
            "Reference file has a larger wavelength grid than CIA file. Will fill with zeros."
        )
    return


def initialize_species_dict(species_dict, wavelength_grid):
    """
    Initializes a dictionary of species to be interpolated.
    Inputs
    ------
        :species_dict: (dict) dictionary of species to be interpolated.
    Outputs
    -------
        None
    Side effects
    -------------
        Modifies species_dict in place.
    """
    for species_key in species_dict.keys():
        species_dict[species_key] += list(np.ones_like(wavelength_grid) * 0.0)
    return species_dict


def add_line_string_species(line_string, species_dict, i, buffer):
    """
    Adds the species to the line string.
    Inputs
    ------
        :line_string: (str) the line string to be modified.
        :species_dict: (dict) dictionary of species to be interpolated.
        :i: (int) index of the line string.
        :buffer: (str) buffer between each species.
    Outputs
    -------
        None
    Side effects
    -------------
        Modifies line_string in place.
    """
    for species_key in species_dict.keys():
        line_string += "{:.12e}".format(species_dict[species_key][i]) + buffer
    return line_string


def build_new_string(
    interped_wavelengths, interped_temps, reference_species, species_dict_interped
):
    """
    Builds the new string that will be written to the new CIA file.
    Inputs
    ------
        :interped_wavelengths: (list) list of wavelengths that have been interpolated.
        :interped_temps: (list) list of temperatures that have been interpolated.
        :reference_species: (list) list of species that have been interpolated.
        :buffer: (str) buffer between each species.
    Outputs
    -------
        :new_string: (list) list of strings that will be written to the new CIA file.
    Side effects
    -------------
        None
    """
    new_string = []
    buffer = "   "  # there's a set of spaces between each string!
    temp = np.nan  # fill value

    for i in tqdm(range(len(reference_species))):
        # the first line gets different treatment!
        if i == 0:
            temp = np.min(
                interped_temps
            )  # add the LOWEST temperature in the temperature grid!
            new_string += ["{:.12e}".format(temp) + "\n"]
        elif interped_temps[i] != temp:
            temp = interped_temps[i]
            new_string += ["{:.12e}".format(temp) + "\n"]
        wavelength_string = "{:.12e}".format(interped_wavelengths[i])

        line_string = wavelength_string + buffer

        line_string = add_line_string_species(
            line_string, species_dict_interped, i, buffer
        )

        new_string += [line_string + "\n"]

    return new_string


def interpolate_cia(
    cia_file, reference_file, outfile=None, loader="exotransmit", load_kwargs=None
):
    """
    Interpolates a CIA file to a higher resolution, using the wavelength grid
    of a reference file. Note: This function assumes that the CIA file has
    Hels, HeHs, CH4CH4s, H2Hes, H2CH4s, H2Hs, H2H2s, and CO2CO2s.

    TODO: add 2d interpolation.
    Inputs
    ------
        :CIA_file: (str) path to CIA file to be interpolated. e.g.,
                    'opacCIA/opacCIA.dat'
        :reference_file: (str) path to opacity file with the wavelength grid of interest. e.g.,
                    'opacFe/opacFe.dat'
    Outputs
    -------
        None
    Side effects
    -------------
        Creates a file with "hires" attached to the end of CIA_file that has been interpolated
        to higher resolution.
    """
    if load_kwargs is None:
        load_kwargs = {}
    reference_opac = Opac(reference_file, loader=loader, load_kwargs=load_kwargs)

    real_wavelength_grid = reference_opac.wl

    # need to put it on the right temperature grid, too!
    real_temperature_grid = reference_opac.T

    df = Opac_cia(cia_file, loader="exotransmit_cia", view="full_frame").cross_section

    check_temp_grid(df, real_temperature_grid, cia_file)

    check_wav_grid(reference_opac, df)

    species_dict_interped = get_empty_species_dict(cia_file)
    interped_wavelengths = []
    interped_temps = []

    # perform interpolation. todo: is this redudnant?
    for unique_temp in tqdm(real_temperature_grid, desc="Interpolating"):
        if unique_temp not in df.temp.unique():
            species_dict_interped = initialize_species_dict(
                species_dict_interped, real_wavelength_grid
            )

        else:
            sub_df = df[df.temp == unique_temp]

            # convert to lists so that they're neatly nested lists
            for species_key in species_dict_interped.keys():
                species_dict_interped[species_key] += list(
                    np.interp(real_wavelength_grid, sub_df.wav, sub_df[species_key])
                )

        interped_wavelengths += list(real_wavelength_grid)
        interped_temps += list(np.ones_like(real_wavelength_grid) * unique_temp)

    # write to a new file
    # this is pretty gross below : (
    reference_species = species_dict_interped[list(species_dict_interped.keys())[0]]

    new_string = build_new_string(
        interped_wavelengths, interped_temps, reference_species, species_dict_interped
    )

    # todo: check the insert. and can pull wavelength grid.
    temp_string = " ".join(str(temp) for temp in real_temperature_grid) + " \n"
    new_string.insert(0, temp_string)
    if isinstance(outfile, type(None)):
        outfile = cia_file.split(".dat")[0]
        outfile += "_highres.dat"
    f2 = open(outfile, "w")
    f2.writelines(new_string)
    f2.close()

    return


# Pt. 2: Chunking


def chunk_wavelengths_cia(file, ref_file_base, numfiles):
    """
    Performs chunking based on the reference file's wavelength chunking.
    Inputs
    -------
        :file: (str) path to CIA file that should be chunked. e.g., opacCIA_highres.dat.
        :ref_file_base: (str) path base to a set of reference files that are already chunked
                        on the desired wavelength grid. e.g., ../opacFe/opacFe
    """

    header = get_header(file)

    # now get chunks

    f = open(file)
    f1 = f.readlines()
    f.close()

    ticker = 0
    file_suffix = 0

    # determine how many wavelengths are in a given chunk
    if "chunk_list.txt" not in os.listdir():
        chunk_list = []
        for i in tqdm(range(numfiles), desc="Putting together chunk list"):
            chunk_list += [get_wav_per_chunk(i, ref_file_base)]

        np.savetxt("chunk_list.txt", chunk_list)

    else:
        chunk_list = np.loadtxt("chunk_list.txt")

    wav_per_chunk = chunk_list[0]  # to start

    ntemps = 0

    temperature = np.nan  # fill value
    for line in tqdm(
        f1, desc="Writing lines to chunk files"
    ):  # read through all lines in the opacity file
        if not line or line == "\n":
            continue  # don't want it to break

        if len(line.split(" ")) == 1:  # this is a new temperature
            temperature = line
            ntemps += 1
            continue  # nothing else on this line
        ticker += 1
        if ticker == wav_per_chunk:
            file_suffix += 1  # start writing to different file
            ticker = 0
            true_file_suffix = file_suffix % numfiles
            wav_per_chunk = chunk_list[true_file_suffix]

            write_to_file(temperature, file, file_suffix, numfiles),
        write_to_file(line, file, file_suffix, numfiles)  # just writes line

    true_header = header[0]

    for chunk_file in glob(file[:-4] + "*"):
        if chunk_file != file:
            prepend_line(chunk_file, true_header)
    return


def get_wav_per_chunk(file_suffix, ref_file_base):
    """
    Grabs the number of wavelengths of a given chunk.
    Inputs
    ------
        :file_suffix: (int) number corresponding to the given chunk. e.g., 1.
        :ref_file_base: (str) path base to a set of reference files that are already chunked
                        on the desired wavelength grid. e.g., ../opacFe/opacFe
    Outputs
    --------
        :len_grid: (int) number of wavelengths in chunk.
    """

    file = ref_file_base + str(file_suffix) + ".dat"

    real_wavelength_grid = get_lams(file, filetype="opacity")

    len_grid = len(real_wavelength_grid)

    return len_grid


def prepend_line(file_name, line):
    """
    Insert given string as a new line at the beginning of a file. Used to fix the header within
    the chunked CIA files.

    Inputs
    -------
        :file_name: (str) path to file to be modified.
        :line: (str) line to be inserted.

    Outputs
    --------
        None

    Side effects
    ------------
        Modifies the file in place.
    """

    # define name of temporary dummy file
    dummy_file = file_name + ".bak"
    # open original file in read mode and dummy file in write mode
    with open(file_name, "r") as read_obj, open(dummy_file, "w") as write_obj:
        # Write given line to the dummy file
        write_obj.write(line + "\n")
        # Read lines from original file one by one and append them to the dummy file
        for line in read_obj:
            write_obj.write(line)

    # remove original file
    os.remove(file_name)
    # Rename dummy file as the original file
    os.rename(dummy_file, file_name)



"""
The high-level API for fitting. Requires the Opac object.
"""
import warnings
from functools import partial
from multiprocessing import Pool

import numpy as np
from tqdm import tqdm

from cortecs.fit.fit_neural_net import *
from cortecs.fit.fit_pca import *
from cortecs.fit.fit_polynomial import *


class Fitter(object):
    """
    fits the opacity data to a neural network.

    todo: fit CIA. only on one dimension, because there's no pressure dependence.
    """

    # make the dictionary of stuff here
    method_dict = {
        "pca": fit_pca,
        "neural_net": fit_neural_net,
        "polynomial": fit_polynomial,
    }

    prep_method_dict = {
        "pca": prep_pca,
        "neural_net": prep_neural_net,
        "polynomial": prep_polynomial,
    }

    save_method_dict = {
        "pca": save_pca,
        "neural_net": save_neural_net,
        "polynomial": save_polynomial,
    }

    def __init__(self, opac, method="pca", **fitter_kwargs):
        """
        todo: make list of opac
        fits the opacity data to a neural network.

        Inputs
        ------
        opac: Opac
            the opacity object.
        method: str
            the method to use for fitting. Options include (in order of increasing complexity) include
            'polynomial', 'pca', and 'neural_net'. The more complex the model, the larger the model size (i.e., potentially
            the lower the compression factor), and the more likely it is to fit well.
        fitter_kwargs: dict
            kwargs that are passed to the fitter. one kwarg, for instance, is the fit_axis: for PCA,
            this determines what axis is fit against.
        """
        self.opac = opac
        self.fitter_kwargs = fitter_kwargs

        if method not in self.method_dict.keys():
            raise ValueError("method {} not supported".format(method))

        self.method = method

        self.fit_func = self.method_dict[self.method]

        self.wl = self.opac.wl
        self.P = self.opac.P
        self.T = self.opac.T

        # todo: figure out how to change the fitting...based on the fit axis?
        return

    def fit(self, parallel=False, verbose=1):
        """
        fits the opacity data to a neural network. loops over all wavelengths.
        Can loop over a list of species? ...should be parallelized!

        Inputs
        ------
        parallel: bool
            whether to parallelize the fitting.
        """

        # iterate over the wavelengths.
        prep_method = self.prep_method_dict[self.method]
        self.prep_res = prep_method(self.opac.cross_section, **self.fitter_kwargs)
        if not parallel:
            self.fit_serial(verbose=verbose)
        else:
            self.fit_parallel()

        return  # will I need to save and stuff like that?

    def fit_serial(self, verbose=0):
        """
        fits the opacity data with a given method. Serial.
        :return:
        todo: keep in mind that the PCA method is not actually independent of wavelength.
        """

        # loop over the wavelengths and fit
        res = []
        with warnings.catch_warnings():
            if verbose == 1:
                iterator = tqdm(enumerate(self.wl), total=len(self.wl))
            else:
                iterator = enumerate(self.wl)
            for i, w in iterator:
                cross_section = self.opac.cross_section[:, :, i]
                res += [
                    self.fit_func(
                        cross_section,
                        self.P,
                        self.T,
                        self.prep_res,
                        **self.fitter_kwargs
                    )
                ]

        self.fitter_results = [self.prep_res, np.array(res)]

        return

    def update_pbar(self, arg):
        """
        Updates a tqdm progress bar.
        """
        self.pbar.update(1)
        pass

    def fit_parallel(self):
        """
        fits the opacity data with a given method. Parallel.
        :return:
        """
        with warnings.catch_warnings():
            num_processes = 1

            func = partial(
                self.fit_func,
                P=self.P,
                T=self.T,
                prep_res=self.prep_res,
                **self.fitter_kwargs
            )

            self.pbar = tqdm(
                total=len(self.wl),
                position=0,
                leave=True,
                unit="chunk",
                desc="Fitting with {} method".format(self.method),
            )

            # these two lines are where the bulk of the multiprocessing happens
            pool = Pool(num_processes)

            # actualy loop over using pool.map. need
            # reformat the cross_section to be a list of 2D arrays
            cross_section_reformatted = [
                self.opac.cross_section[:, :, i] for i in range(len(self.wl))
            ]

            # we tehcnically want sorted results. but apply async is the only way to get the progress bar to work!
            async_result = []
            for i, item in enumerate(cross_section_reformatted):
                async_result.append(
                    [i, pool.apply_async(func, args=(item,), callback=self.update_pbar)]
                )

            # Close the pool
            pool.close()
            pool.join()

            # Close the progress bar
            self.pbar.close()

            # sort the results based on the index
            sorted_results = [None] * len(async_result)
            for item in async_result:
                i, res = item
                sorted_results[i] = res.get()

        self.fitter_results = [self.prep_res, sorted_results]

        return

    def save(self, savename):
        """
        saves the fitter results.
        """
        save_func = self.save_method_dict[self.method]
        save_func(savename, self.fitter_results)
        return



"""
Module for performing PCA regression on opacity functions.
"""

import numpy as np


def standardize_cube(input_array):
    """
    Prepares an array for PCA.

    Inputs
    -------
        :input_array: (ntemperature x npressure) array for PCA.

    Returns
    -------
        :standardized_cube: (ntemperature x npressure) standardized (mean-substracted, standard deviation-divided) array for PCA.
    """
    nf, nx = input_array.shape
    mat = input_array.copy()

    if np.any(np.isnan(mat)):
        raise ValueError("NaNs in input array.")
    # loop over the first index
    for i in range(nf):
        mat[i, :] -= np.mean(mat[i, :])
        mat[i, :] /= np.std(mat[i, :])
    standardized_cube = mat
    return standardized_cube


def do_svd(standardized_cube, nc, nx):
    """
    Does the SVD (singular value decomposition) on the standardized cube.

    Inputs
    -------
        :standardized_cube: (ntemperature x npressure) standardized (mean-substracted, standard deviation-divided) array for PCA.
        :nc: (int) number of PCA components to keep in the reconstruction. Increasing the number of components can make the reconstruction
        of opacity data more accurate, but it can also lead to overfitting and make the model size larger (i.e.,
        decrease the compression factor).

    """

    xMat = np.ones((nx, nc + 1))  # Note the extra column of 1s
    mat = np.moveaxis(standardized_cube.copy(), 0, -1)
    u, s, vh = np.linalg.svd(mat, full_matrices=False)
    xMat[:, 1:] = u[:, 0:nc]  # The first column is left as a column of ones

    return xMat, s, vh, u


def fit_mlr(cube, X):
    """
    todo: check array shapes.
    Fits the MLR (multiple linear regression) to the input cube with the PCA outputs and normalizes accordingly.

    Inputs
    -------
        :cube: (ntemp x npressure) array being fit with PCA. Generally, this is the opacity data at a single
        wavelength as a function of temperature and pressure.
        :X: (npressure x ncomponents) PCA components. these are set by functions such as `do_pca`.

    Returns
    -------
        :beta: (ncomponents x ntemp) PCA coefficients. These are the coefficients that are used to reconstruct the
        opacity data at a later stage (e.g., within an atmospheric forward model). They represent the "compressed" opacity.
    """
    Y = np.moveaxis(cube.copy(), 0, -1)
    term1 = np.linalg.inv(np.dot(X.T, X))
    term2 = np.dot(term1, X.T)
    beta = np.dot(term2, Y)
    return beta


def do_pca(cube, nc=3):
    """
    Does all the PCA steps. right now for a single spectrum.

    Inputs
    -------
        :cube: (ntemperature x npressure) array being fit with PCA. Generally, this is the opacity data at a single
        wavelength as a function of temperature and pressure.
        :nc: (int) number of PCA components to keep. Increasing the number of components can make the reconstruction
        of opacity data more accurate, but it can also lead to overfitting and make the model size larger (i.e.,
        decrease the compression factor).

    """

    # do the PCA
    standardized_cube = standardize_cube(cube)

    nf, nx = cube.shape
    """ Choosing the number of components and initialising the matrix of eigenvectors """

    try:
        xMat, s, vh, u = do_svd(standardized_cube, nc, nx)

    except np.linalg.LinAlgError as e:
        print("SVD did not converge.")
        raise e

    return xMat, standardized_cube, s, vh, u


def fit_pca(cross_section, P, T, prep_res, fit_axis="pressure", **kwargs):
    """
    Fits the PCA to the opacity data.

    Inputs
    -------
        :cross_section: (ntemp x npressure) the array of cross-sections being fit.
        :P: pressure grid
        :T: temperature grid
        :xMat: (npres x nc) PCA components

    Returns
    -------
        :beta: (nc x pixels) PCA coefficients
    """
    # print("shapes for everything:", cross_section.shape, P.shape, T.shape, xMat.shape)
    xMat = prep_res
    cross_section = move_cross_section_axis(cross_section, fit_axis)
    beta = fit_mlr(cross_section, xMat)
    return beta


def move_cross_section_axis(cross_section, fit_axis, dim=2):
    """
    todo: add docstring
    :param cross_section:
    :param fit_axis:
    :return:
    """
    fit_axis_options = ["best", "temperature", "pressure"]
    if fit_axis not in fit_axis_options:
        raise ValueError(f"fit_axis param must be one of: {fit_axis_options}")

    # the current shape prefers pressure. let's auto-check, though
    if fit_axis == "best":
        # actually want SECOND longest.
        if dim == 3:
            longest_axis = np.argmax(cross_section[:, :, 0].shape)
        else:
            longest_axis = np.argmax(cross_section.shape)

        # now move that longest axis to 0.
        cross_section = np.moveaxis(cross_section, longest_axis, 1)

    elif fit_axis == "temperature":
        cross_section = np.moveaxis(cross_section, 0, 1)

    return cross_section


def prep_pca(
    cross_section, wav_ind=-1, nc=2, force_fit_constant=False, fit_axis="pressure"
):
    """
    Prepares the opacity data for PCA. That is, it calculates the PCA components to be fit along the entire
    dataset by fitting the PCA to a single wavelength.

    todo: perform type checking for inputs.

    Inputs
    -------
        :cross_section: (ntemp x npressure x nwavelength) the array of cross-sections being fit.
        :wav_ind: (int) index of wavelength to fit. Ideally, this should be fit to a wavelength that has somewhat
        representative temperature--pressure structure to the opacity function. If there is *no* temperature or wavelength
        dependence at this wavelength index but strong dependence elsewhere, for instance, the opacity function will
        be poorly reconstructed.
        :nc: (int) number of PCA components to keep. Increasing the number of components can make the reconstruction
        of opacity data more accurate, but it can also lead to overfitting and make the model size larger (i.e.,
        decrease the compression factor).
        :force_fit_constant: (bool) if True, will allow the PCA to fit an opacity function without temperature and pressure
        dependence. This usually isn't recommended, if these PCA vectors are to e used to fit other wavelengths that
        *do* have temperature and pressure dependence.
        :fit_axis: (str) the axis to fit against. determines the shape of the final vectors and components.
        if "best", chooses the largest axis. otherwise, can select "temperature" or "pressure".

    Returns
    -------
        :xMat: (n_exp x nc) PCA components
    """
    cross_section = move_cross_section_axis(cross_section, fit_axis, dim=3)
    single_pres_single_temp = cross_section[:, :, wav_ind]
    if (
        np.all(single_pres_single_temp == single_pres_single_temp[0, 0])
        and not force_fit_constant
    ):
        raise ValueError(
            "all values are the same at this wavelength index! Try a different one."
        )

    xMat, fStd, s, vh, u = do_pca(single_pres_single_temp, nc)
    return xMat


def save_pca(savename, fit_results):
    """
    Saves the PCA components and coefficients to files

    Inputs
    -------
        :savename: (str) if not None, the PCA components will be saved to this filename.
        :fit_results: contains the PCA coeefficients and vectors.
    """
    # pdb.set_trace()
    vectors, beta = fit_results
    np.savez_compressed(savename, pca_coeffs=beta, vectors=vectors)



"""
this file holds the class describing opacity data. hmm...maybe I want the loader in a different file?

author: @arjunsavel
"""
import numpy as np

from cortecs.opac.io import *


class Opac(object):
    """
    this class holds the opacity data and provides methods for evaluating the opacity at a given temperature and pressure.

    Everything's already been loaded into memory, so this is just a wrapper for the data.

    todo: use the different loaders
    """

    method_dict = {
        "chimera": loader_chimera,
        "helios": loader_helios,
        "platon": loader_platon,
        "exotransmit": loader_exotransmit,
    }

    wl = None
    T = None
    P = None

    def __init__(
        self,
        filename,
        loader="chimera",
        load_kwargs={},
    ):
        """
        wraps around the loaders.

        Parameters
        ----------
        filename : str
            name of file to load
        loader : str
            name of loader to use. default is chimera.

        Returns
        -------
        nothing
        """
        self.filename = filename
        self.load_obj = self._get_loader(loader)
        self.wl, self.T, self.P, self.cross_section = self.load_obj.load(
            filename, **load_kwargs
        )

        self.n_wav, self.n_t, self.n_p = len(self.wl), len(self.T), len(self.P)

        return

    def _get_loader(self, loader_name):
        """
        gets the loader object

        Parameters
        ----------
        loader_name : str
            name of loader

        Returns
        -------
        loader object
        """
        if loader_name not in self.method_dict:
            raise ValueError(
                "loader name not found. valid loaders are: {}".format(
                    self.method_dict.keys()
                )
            )
        return self.method_dict[loader_name]()

    # todo: implement the copy and deepcopy methods.
    def copy(self):
        # need a new opac object with all the same attributes. so let's make sure we don't load on input.
        raise NotImplementedError


class Opac_cia(Opac):
    """
    this class holds the opacity data and provides methods for evaluating the opacity at a given temperature and pressure.

    Everything's already been loaded into memory, so this is just a wrapper for the data.

    """

    method_dict = {
        "platon_cia": loader_platon_cia,
        "exotransmit_cia": loader_exotransmit_cia,
    }

    def __init__(self, filename, loader="platon_cia", view="default"):
        """
        wraps around the loaders.
        NOTE: there's a mixed type for self.cross_section. can be a dataframe if there are a number of species.

        Parameters
        ----------
        filename : str
            name of file to load
        loader : str
            name of loader to use. default is chimera.
        view : str
            if 'full_frame' is selected, the cross_section object also includes temperature and wavelength columns.

        Returns
        -------
        nothing
        """
        self.filename = filename
        self.load_obj = self._get_loader(loader)
        self.wl, self.T, self.cross_section = self.load_obj.load(self.filename)

        # at least for the exotransmit case, we have...wl x temp.
        if view == "full_frame":
            self.cross_section["temp"] = self.T
            self.cross_section["wav"] = self.wl

        self.n_wav, self.n_t = len(self.wl), len(self.T)

        return

    def join_cross_section(self, opac):
        """
        joins another opacity's cross-section data to this one.
        :param opac:
        :return:
        """

        def replace_zeros(species):
            mask = self.cross_section[species] == 0.0
            self.cross_section.loc[mask, species] = opac.cross_section.loc[
                mask, species
            ]

        # throw errors if the grids don't match.
        if np.all(opac.T != self.T):
            raise ValueError("temperatures do not match!")
        if np.all(opac.wl != self.wl):
            raise ValueError("wavelengths do not match!")

        # add the entirely new columns
        other_columns = opac.cross_section.columns
        for column in other_columns:
            if column not in self.cross_section.columns:
                self.cross_section[column] = opac.cross_section[column]

        # now fill in any temperatures where the cross-section is zero here.
        for species in self.cross_section.columns:
            if species in other_columns:
                replace_zeros(species)

        return



# Inspired by:
# https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/

import codecs
import os
import re

from setuptools import find_packages, setup

###################################################################

NAME = "cortecs"
PACKAGES = find_packages(where="src")
META_PATH = os.path.join("src", NAME, "__init__.py")
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Topic :: Scientific/Engineering :: Astronomy",
]
# todo: pin to requirements
INSTALL_REQUIRES = [
    "numpy",
    "jaxlib",
    "jax",
    "tqdm",
    "h5py",
    "matplotlib",
    "pandas",
]

###################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


if __name__ == "__main__":
    setup(
        name=NAME,
        description=find_meta("description"),
        license=find_meta("license"),
        url=find_meta("uri"),
        version=find_meta("version"),
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        package_data={"": ["README.md", "LICENSE"]},
        long_description=read("README.md"),
        long_description_content_type="text/markdown",
        packages=PACKAGES,
        package_dir={"": "src"},
        zip_safe=False,
        python_requires=">3.8.0",
        classifiers=CLASSIFIERS,
        include_package_data=True,
        extras_require={
            "neural_networks": [
                "tensorflow",
                "keras",
            ]
        },
        install_requires=INSTALL_REQUIRES,
        options={"bdist_wheel": {"universal": "1"}},
    )



"""
Plots histograms and stuff after the fitting has occurred so that the level of fit can be assessed.
"""
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm

from cortecs.eval.eval import Evaluator


def plot_loss(history):
    """
    plots the loss from the history object. need to substantially rewrite
    TODO: rcparams file to make things look nice, fontwise and such?
    :param history:
    :return:
    """
    plt.plot(history.history["loss"], label="loss", color="teal", lw=3)
    plt.xlabel("Epoch", fontsize=20)
    plt.ylabel("Error", fontsize=20)  # need the units on this
    plt.yscale("log")
    plt.legend()
    return


def title_format(quantity):
    """
    formats the title of the histogram based on the quantity being plotted.

    Inputs
    ------
    quantity: array-like
        the quantity being plotted.
    """
    plt.title(
        "Mean: {:.2f}, Std: {:.2f}".format(np.mean(quantity), np.std(quantity)),
        fontsize=20,
    )
    return


def calc_metrics(fitter, tp_undersample_factor=1, wl_under_sample_factor=8, plot=False):
    """
        calculates the mean and percent error  associated with the fit.
    .
        Inputs
        ------
        opac_obj: Opac
            the opacity object to be assessed.
        fitter: Fitter
            the fitter object that was used to fit the opacity object.
        tp_undersample_factor: int
            factor to undersample the temperature and pressure grids by when calculating the fit metrics
        wl_under_sample_factor: int
            factor to undersample the wavelength grid by.
    """
    evaluator = Evaluator(fitter.opac, fitter)
    AMU = 1.6605390666e-24  # atomic mass unit in cgs. From astropy!

    try:
        species_weight = evaluator.load_obj.species_weight
    except AttributeError:
        raise AttributeError(
            "Object has not been loaded to include the species weight. Please load the opacities appropriately."
        )

    vals = []
    orig_vals = []
    abs_diffs = []
    percent_diffs = []
    for i in tqdm(range(len(fitter.opac.T))[::tp_undersample_factor]):
        for j in range(len(fitter.opac.P))[::tp_undersample_factor]:
            for k in range(len(fitter.opac.wl))[::wl_under_sample_factor]:
                val = evaluator.eval(
                    fitter.opac.T[i], fitter.opac.P[j], fitter.opac.wl[k]
                )
                # todo: check that this works for not just PLATON
                val = np.log10(val * species_weight * AMU * 1e-4)
                if not np.isfinite(val):
                    val = -104
                orig_val = fitter.opac.cross_section[i, j, k]
                abs_diffs += [val - orig_val]
                percent_diffs += [100 * (val - orig_val) / orig_val]
                vals += [val]
                orig_vals += [orig_val]
    abs_diffs = np.array(abs_diffs)
    percent_diffs = np.array(percent_diffs)
    orig_vals = np.array(orig_vals)
    vals = np.array(vals)

    if plot:
        plt.figure()
        plt.hist(abs_diffs, color="dodgerblue")
        plt.xlabel("Abs. residuals in log10 opacity", fontsize=20)
        plt.ylabel("Count", fontsize=20)

        title_format(abs_diffs)

        plt.figure()
        plt.hist(percent_diffs, color="goldenrod")
        plt.xlabel("Percent. residuals in log10 opacity", fontsize=20)
        title_format(abs_diffs)

        plt.ylabel("Count", fontsize=20)

    return vals, orig_vals, abs_diffs, percent_diffs



"""
For evaluating the PCA fit at a given temperature, pressure, and wavelength.
"""

import jax
import numpy as np


@jax.jit
def eval_pca_ind_wav(first_ind, second_ind, vectors, pca_coeffs):
    """
    Evaluates the PCA fit at a given temperature and pressure.

    Unfortunately, not all GPUs will support a simpler dot product, I believe,
    Also, we cannot loop over n_components explicitly because JAX
    functions require static loop lengths.

    Inputs
    ------
    first_ind: int
        The index of the first axis quantity (default temperature) to evaluate at.
    second_ind: int
        The index of the second axis quantity (default pressure) to evaluate at.
    vectors: array
        The PCA vectors.
    pca_coeffs: array
        The PCA coefficients.
    n_components: int
        The number of PCA components used in the fitting


    """

    xsec_val = 0.0
    n_components = vectors.shape[1]
    for component in range(n_components):
        xsec_val += vectors[first_ind, component] * pca_coeffs[component, second_ind]
    return xsec_val


def eval_pca(
    temperature,
    pressure,
    wavelength,
    T,
    P,
    wl,
    fitter_results,
    fit_axis="pressure",
    **kwargs
):
    """
    Evaluates the PCA fit at a given temperature, pressure, and wavelength.


    Inputs
    ------
    temperature: float
        The temperature to evaluate at.
    pressure: float
        The pressure to evaluate at.
    wavelength: float
        The wavelength to evaluate at.

    """
    # find the nearest temperature, pressure, and wavelength indices.

    temperature_ind = np.argmin(np.abs(T - temperature))
    pressure_ind = np.argmin(np.abs(P - pressure))
    wavelength_ind = np.argmin(np.abs(wl - wavelength))

    pca_vectors, pca_coeffs_all_wl = fitter_results
    pca_coeffs = pca_coeffs_all_wl[wavelength_ind, :, :]

    # todo: figure out how to order the pressure and temperature inds!
    # pdb.set_trace()
    if fit_axis == "pressure":
        first_arg = pressure_ind
        second_arg = temperature_ind
    elif fit_axis == "temperature":
        first_arg = temperature_ind
        second_arg = pressure_ind

    elif fit_axis == "best":
        T_length = len(T)
        P_length = len(P)

        # todo: what if equal?
        if T_length > P_length:
            first_arg = temperature_ind
            second_arg = pressure_ind
        else:
            first_arg = pressure_ind
            second_arg = temperature_ind
    # print("first_arg, second_arg", first_arg, second_arg)
    # print("shapes:", pca_vectors.shape, pca_coeffs.shape)
    return eval_pca_ind_wav(first_arg, second_arg, pca_vectors, pca_coeffs)
